"""
Droplet Contact Numbers and Probabilities (DCNumPro)

Determine droplet contact numbers and probabilities based on 
the median and the standard deviation of a droplet size distribution.

Source code of DCNumPro: https://git.rwth-aachen.de/avt-fvt/public/contact-numbers-and-probabilities/
DCNumPro uses a 3D cell-based Voronoi library based on voro++: https://github.com/wackywendell/tess

The source code of DCNumPro is licensed under the Eclipse Public License v2.0 (EPL 2.0).
Copyright (c) 2024 Fluid Process Engineering (AVT.FVT), RWTH Aachen University.
Written by Stepan Sibirtsev & Andrey Kirsanov

The coyprights and license terms are given in LICENSE.

This script contains all functions related to droplet size distribution (DSD).
Script version date: 23.11.2024
"""  

""" Import packages """
import random
import numpy as np
from scipy.stats import norm
from scipy.stats import lognorm

""" Functions """
def calc_distribution(config):
    """ Generate nuber distribution of particles based on the defined distribution parameters mu and sigma """
    # Case differentiation: DSD type
    if config.dsd_type == 0:
        # CDF parameter for log-normal DSD
        # Mean of log values
        mu_log = (np.log((config.mu**2)/(((config.sigma**2)+(config.mu**2))**0.5)))
        # Standard deviation of log values
        sigma_log = ((np.log(((config.sigma**2)/(config.mu**2))+1))**0.5)
        # DSD boundaries which are d_05 and d_95 corresponding to the 5th and 95th percentile of the DSD 
        d_05, d_95 = calc_DSD_boundaries(config, mu_log, sigma_log)
    elif config.dsd_type == 1:
        # DSD boundaries which are d_05 and d_95 corresponding to the 5th and 95th percentile of the DSD 
        d_05, d_95 = calc_DSD_boundaries(config, config.mu, config.sigma)
    # Class interval
    delta_class = (d_95-d_05)/config.n_classes
    # Class diameters
    d_classes = np.linspace((d_05+delta_class/2), (d_95-delta_class/2), config.n_classes)
    # Determine minimum and maximum class diameter
    d_min = np.min(d_classes)
    d_max = np.max(d_classes)
    # Class boundaries
    boundaries = [d_05 + i * delta_class for i in range(config.n_classes+1)]
    # Class sizes
    # Case differentiation: DSD type
    if config.dsd_type == 0:
        class_sizes, q0_initial, q3_initial, volume_total = calc_class_sizes(config, mu_log, sigma_log, boundaries, d_classes)
    elif config.dsd_type == 1:
        class_sizes, q0_initial, q3_initial, volume_total = calc_class_sizes(config, config.mu, config.sigma, boundaries, d_classes)
    # Sauter mean diameter
    D_32 = np.sum(q0_initial * d_classes**3) / np.sum(q0_initial * d_classes**2)
    # Actual number of particles
    n_particles_init = int(np.sum(class_sizes))
    # Create array q0 [class_sizes, d_classes]
    q0 = np.concatenate((class_sizes.reshape(-1,1), d_classes.reshape(-1,1)), axis=1)

    return q0, n_particles_init, q0_initial, q3_initial, d_classes, d_min, d_max, volume_total, d_05, d_95, delta_class, D_32

def calc_DSD_boundaries(config, mu, sigma): 
    """ Calculate DSD boundaries d_05 and d_95 corresponding to the 5th and 95th percentile of the DSD. """
    # Find the z-value corresponding to the 5th and 95th percentile of the standard normal distribution
    d_05_percentile = norm.ppf(0.05)
    d_95_percentile = norm.ppf(0.95)
    # Case differentiation: DSD type
    if config.dsd_type == 0:
        # Calculate the 5th and 95th percentile of the log-normal distribution
        d_05 = np.exp(mu + sigma * d_05_percentile)
        d_95 = np.exp(mu + sigma * d_95_percentile)
    elif config.dsd_type == 1:
        d_05 = mu + sigma * d_05_percentile
        d_95 = mu + sigma * d_95_percentile

    return d_05, d_95

def calc_class_sizes(config, mu, sigma, boundaries, d_classes):
    """ Calculate class sizes for a log-normal distribution with parameters mu and sigma. """
    # Ensure boundaries are sorted
    boundaries = sorted(boundaries)
    # Case differentiation: DSD type
    if config.dsd_type == 0:
        # Create a log-normal distribution object
        # Scale parameter for lognorm is sigma
        # Scale parameter exp(mu) for shifting the distribution
        dist = lognorm(s=sigma, scale=np.exp(mu))
    elif config.dsd_type == 1:
        # Create a normal distribution object
        dist = norm(loc=mu, scale=sigma)
    # Calculate probabilities for each interval
    q3_initial = []
    for i in range(len(boundaries) - 1):
        lower_bound = boundaries[i]
        upper_bound = boundaries[i + 1]
        # Probability that a value falls within the interval [lower_bound, upper_bound]
        prob = dist.cdf(upper_bound) - dist.cdf(lower_bound)
        # Calculate class sizes
        q3_initial.append(prob)
    q3_initial = np.array(q3_initial)
    # Convert class volume probabilities into number probabilities
    q0_initial = q3_initial / d_classes**3
    # Normalize number probabilities. 
    # It ignores that the class volume probabilities include only 90vol% of the distribution.
    # However, since we use number probabilities to distribute a certain number of particles 
    # within these considered 90vol%, this calculation is valid
    q0_initial /= np.sum(q0_initial)
    # Distribute the nubmer of particles to the classes considering number probabilities and round
    class_sizes = config.n_particles_sim * np.array(q0_initial)
    np.round(class_sizes, decimals=0, out=class_sizes)
    # Total particle volume
    volume_total = np.sum(class_sizes * (np.pi * d_classes**3 / 6))

    return class_sizes, q0_initial, q3_initial, volume_total

def calc_particles(config, q0, n_particles_init):
    """
    Function to fill an array with particles from number distribution q0 and determine their drop coordinates randomly.
    Args:
    q0 (array): Number distribution.
    n_particles_init (float): Actualnumber of particles.
    Returns:
    M_particles_initial (array): Array with particles and their drop coordinates
    """
    # Array columns: 
    # 0: x-coordinate, 1: y-coordinate, 2: z-coordinate, 3: index, 4: diameter, 5: class number (0 to n)
    M_particles_initial = np.zeros(shape=(int(n_particles_init), 6)) 
    # Assign particle diameters and class numbers
    j = 0
    for k in range(len(q0)):
        n = int(q0[k, 0])
        for i in range(n):
            M_particles_initial[j, 4] = q0[k, 1]
            M_particles_initial[j, 5] = k
            j = j + 1
    # Assign indices and randomly generated coordinates
    for i in range(len(M_particles_initial)):
        M_particles_initial[i, 0] = random.uniform(-config.A_0 + (M_particles_initial[i, 4]) / 2, config.A_0 - (M_particles_initial[i, 4]) / 2)
        M_particles_initial[i, 1] = random.uniform(-config.A_0 + (M_particles_initial[i, 4]) / 2, config.A_0 - (M_particles_initial[i, 4]) / 2)
        M_particles_initial[i, 2] = None
        M_particles_initial[i, 3] = i
    np.random.shuffle(M_particles_initial)

    return M_particles_initial

def particle_info(M_particles, index):
    """ This function determines the particle info like coordinates and diameter """
    # x-coordinate
    x = M_particles[index, 0]
    # y-coordinate
    y = M_particles[index, 1]
    # z-coordinate
    z = M_particles[index, 2]
    # Index
    p_index = M_particles[index, 3]
    # Diameter
    p_diameter = M_particles[index, 4]
    # Class
    p_class = int(M_particles[index, 5])
    
    return x, y, z, p_index, p_diameter, p_class

def calc_q(diameters, n_particle, d_classes, delta_class):
    """ Function to calculate the number and volume distributions """
    # Predefine array for class sizes after random loose packing simulation
    class_sizes = np.zeros(len(d_classes), dtype=int)
    # Categorize cell diameters into classes
    for particle in diameters:
        for i in range(len(d_classes)):
            lb_class = d_classes[i] - delta_class/2
            ub_class = d_classes[i] + delta_class/2
            if lb_class <= particle < ub_class:
                class_sizes[i] += 1
                break
    # Final number distribution after growing simulation
    q0 = class_sizes / float(n_particle)
    # Final volume distribution after growing simulation
    # Factor 0.9, since we consider the 90% of the total volume in our methods
    q3 = 0.9*(q0 * (np.pi/6) * d_classes**3) / np.sum(q0 * (np.pi/6) * d_classes**3)

    return q0, q3