"""
Droplet Contact Numbers and Probabilities (DCNumPro)

Determine droplet contact numbers and probabilities based on 
the median and the standard deviation of a droplet size distribution.

Source code of DCNumPro: https://git.rwth-aachen.de/avt-fvt/public/contact-numbers-and-probabilities/
DCNumPro uses a 3D cell-based Voronoi library based on voro++: https://github.com/wackywendell/tess

The source code of DCNumPro is licensed under the Eclipse Public License v2.0 (EPL 2.0).
Copyright (c) 2024 Fluid Process Engineering (AVT.FVT), RWTH Aachen University.
Written by Stepan Sibirtsev & Andrey Kirsanov

The coyprights and license terms are given in LICENSE.

This is the main script to execute DCNumPro.
Script version date: 23.11.2024
"""  

""" Import packages """
import os
import time
import sys
import math
import random
import warnings
import multiprocessing
import numpy as np
from functools import reduce
from scipy.optimize import minimize
from functools import partial
from tqdm import tqdm

""" Import configuration """
import DCNumPro_config as config

""" Import droplet size distribution (DSD) related functions """
# Function to generate nuber distribution of particles based on the defined distribution parameters mu and sigma calculated
from DCNumPro_DSD import calc_distribution
# Function to fill an array with particles from number distribution q0 and determine their drop coordinates randomly.
from DCNumPro_DSD import calc_particles
# This function determines the particle info like coordinates and diameter
from DCNumPro_DSD import particle_info
# Function to calculate the number and volume distributions
from DCNumPro_DSD import calc_q

""" Import random loose packing (RLP) related functions """
# Function / Algorithm to determine the steady state position of the considered particle within the RLP.
from DCNumPro_RLP import calc_position
# This function determines the drop position of particles
from DCNumPro_RLP import find_z_drop

""" Import Tesselation and particle grow related functions """
# Function to perform Tesselation
from DCNumPro_grow import voro_tess
# Function to simulate particle grow
from DCNumPro_grow import calc_grow
# Solver to calculate the grow factor
from DCNumPro_grow import solver_grow_factor

""" Import contact number and probability related functions """
# This function calculates the average contact number of each class and the contact probability between classes.
from DCNumPro_ContProb import calc_contact_probabilities

""" Import data export related functions """
# Function to export results data to excel file
from DCNumPro_excel import save_excel 

""" Import plot related functions """
# Function to graphically display spheres in 3D
from DCNumPro_plot import plot_particles

""" Solver warning messages """ 
if config.display_solver_warnings == False:
    warnings.filterwarnings("ignore")   

""" Assign configuration parameter to global parameter """
# Specify parameter in the .job file  
if config.cluster == True:
    import argparse
    parser = argparse.ArgumentParser(description='Droplet Contact Numbers and Probabilities.')
    parser.add_argument('--name_excel_file', required=True)
    parser.add_argument('--n_particles_sim', required=True, type=int)
    parser.add_argument('--n_classes', required=True, type=int)    
    parser.add_argument('--holdup_est', required=True, type=float)
    parser.add_argument('--seed', required=True, type=int)
    parser.add_argument('--n_cores', required=True, type=int)
    parser.add_argument('--holdup_delta_const', required=True, type=float)
    parser.add_argument('--holdup_delta_nonconst', required=True, type=float)
    parser.add_argument('--stepsize', required=True, type=float)
    parser.add_argument('--mu', required=True, type=float)
    parser.add_argument('--sigma', required=True, type=float)
    args = parser.parse_args()
    config.n_particles_sim = args.n_particles_sim
    config.n_classes = args.n_classes
    config.holdup_est = args.holdup_est
    config.seed = args.seed
    config.n_cores = args.n_cores
    config.holdup_delta_const = args.holdup_delta_const
    config.holdup_delta_nonconst = args.holdup_delta_nonconst
    config.stepsize = args.stepsize
    config.mu = args.mu
    config.sigma = args.sigma
    NAME_EXCEL_FILE = (args.name_excel_file + '.xls')
    ROOT_DIR = os.path.join("/rwthfs/rz/cluster", os.path.abspath("../output"))
else:
    NAME_EXCEL_FILE = (config.name_excel_file + '.xls')
    ROOT_DIR = os.path.abspath("../output")
# Check if the folder exists
if not os.path.exists(ROOT_DIR):
    # Create the folder
    os.makedirs(ROOT_DIR)

""" MAIN """
if __name__ == "__main__":
    # Time stamp
    start_time_rlp = time.time()
    print('Starting simulation with seed', config.seed)
    # Set random seed
    random.seed(config.seed)
    np.random.seed(config.seed)
    sys.stdout.write("Generating particle size distribution...")
    # Number distribution of particles
    q0, n_particles_init, q0_initial, q3_initial, d_classes, d_min, d_max, volume_total, d_05, d_95, delta_class, D_32 = calc_distribution(config)
    # Volume of the simulation environment
    volume_sim = volume_total / config.holdup_est
    # Cuboid's edge lenght (µm)
    L_0 = volume_sim ** (1/3)
    # Half of cuboid's edge lenght (µm)
    config.A_0 = L_0 / 2
    # Mix particles in array and determine their coordinates randomly --> initial array for particles
    M_particles_initial = calc_particles(config, q0, n_particles_init)
    # Array for particles in steady state. 
    # Each row corresponds to one particle.
    # Columns: 0: index of this array, 1: z-coordinate, 2: y-coordinate, 3: x-coordinate, 4: steady state status,
    # 5: particle diameter, 6: index of initial array, 7: particle class 
    # further columns (len(q0)) corresponds to number of classes of the distribution.
    # Thus, contact particle classes can be assigned to each particle in array.
    M_particle_steady = np.zeros((1, (8 + len(q0))))
    sys.stdout.write("completed\n")
    # Use tqdm to track progress
    with tqdm(total=n_particles_init, desc='Random loose packing simulation progress', unit='particle') as pbar:
        # Drop and roll loop:
        for particle_i in range(len(M_particles_initial)):
            if config.display_detailed_progress == True:
                print('Particle Nr.:', particle_i)
            """ Take information of considered particles from initial particle array to define its initial state """
            x, y, z, index_i, d, class_i = particle_info(M_particles_initial, particle_i)
            """ Define test range: the distance between the centers of the considered particle and the bigest particlein steady state"""
            # List of all particles in steady state
            d_steady = M_particle_steady[:, 5] 
            # Diameter of biggest particle in steady state
            d_max_steady = np.max(d_steady) 
            # Test range
            test_range = 1.2*((d + d_max_steady) / 2)
            # The first particle is simply set because there is nothing there before.
            if particle_i == 0:
                # Calculate new z-coordinate
                z = 0 + d / 2
                """Fill steady state array"""
                # Particle index
                M_particle_steady[0, 0] = particle_i
                # z-coordinate
                M_particle_steady[0, 1] = z
                # y-coordinate
                M_particle_steady[0, 2] = y
                # x-coordinate
                M_particle_steady[0, 3] = x
                # Contact with ground, 1=yes
                M_particle_steady[0, 4] = 1
                # Particle diameter
                M_particle_steady[0, 5] = d
                # Initial index
                M_particle_steady[0, 6] = index_i
                # Particle class
                M_particle_steady[0, 7] = class_i
                continue
            # Calculate dropping height based on the location of the heighest particle
            z = find_z_drop(y, x, M_particle_steady, test_range, d_max_steady)
            # Initial dropping position
            z_drop = z
            y_drop = y
            x_drop = x
            # Determine the steady state position of the considered particle within the RLP.
            z, y, x, steady, overlaps, fatal_error = calc_position(config, z, y, x, z_drop, y_drop, x_drop, d, M_particle_steady, test_range, d_max_steady, particle_i)
                
            """ Display final particle status and store the steady state particle in the steady state array. """
            # Display if the considered particle is in steady state
            if config.display_detailed_progress == True:
                if steady == 1: 
                    print('Particle Nr.', particle_i, 'reaches the ground.')
                elif steady == 2:
                    print('Particle Nr.', particle_i, 'has reached steady state on three particles.')
                elif steady == 3:
                    print('Particle Nr.', particle_i, 'has reached steady state on three particles.')
                elif steady == 4:
                    print('Particle Nr.', particle_i, 'has reached steady state on two particles and one wall.')
                elif steady == 5:
                    print('Particle Nr.', particle_i, 'has reached steady state on three particles and one wall.')
                elif steady == 6:
                    print('Particle Nr.', particle_i, 'has reached steady state on one particles and two walls.')
                elif steady == 7:
                    print('Particle Nr.', particle_i, 'has reached steady state on two particles and two walls.')
            # If error=True, timeout in the calc_sedimentation function occured. 
            # The considered particle is stored in the steady state array with NaN values.
            if fatal_error == True:
                M_particle_steady = np.vstack([M_particle_steady, np.zeros((1, 8 + len(q0)))])
                M_particle_steady[len(M_particle_steady) - 1, 0] = particle_i
                M_particle_steady[len(M_particle_steady) - 1, 1] = np.nan
                M_particle_steady[len(M_particle_steady) - 1, 2] = np.nan
                M_particle_steady[len(M_particle_steady) - 1, 3] = np.nan
                M_particle_steady[len(M_particle_steady) - 1, 4] = np.nan
                M_particle_steady[len(M_particle_steady) - 1, 5] = d
                M_particle_steady[len(M_particle_steady) - 1, 6] = index_i
                M_particle_steady[len(M_particle_steady) - 1, 7] = class_i
            # The considered particle is stored in the steady state array.
            else:    
                M_particle_steady = np.vstack([M_particle_steady, np.zeros((1, 8 + len(q0)))])
                M_particle_steady[len(M_particle_steady) - 1, 0] = particle_i
                M_particle_steady[len(M_particle_steady) - 1, 1] = z
                M_particle_steady[len(M_particle_steady) - 1, 2] = y
                M_particle_steady[len(M_particle_steady) - 1, 3] = x
                M_particle_steady[len(M_particle_steady) - 1, 4] = steady
                M_particle_steady[len(M_particle_steady) - 1, 5] = d
                M_particle_steady[len(M_particle_steady) - 1, 6] = index_i
                M_particle_steady[len(M_particle_steady) - 1, 7] = class_i 
                # Store contact partners in the steady state array.
                # Determine number of collisions
                n_overlaps_final = len(overlaps)
                if n_overlaps_final == 1:
                    # Particle classes of the particles in contact
                    class1 = int(M_particle_steady[overlaps[0], 7])
                    # The classes of the particles in contact are assigned to the considered particle
                    M_particle_steady[len(M_particle_steady) - 1, 8 + class1] = M_particle_steady[len(M_particle_steady) - 1, 8 + class1] + 1
                    # The class of the considered particle is assigned to the particles in contact 
                    M_particle_steady[overlaps[0], 8 + class_i] = M_particle_steady[overlaps[0], 8 + class_i] + 1
                elif n_overlaps_final == 2:
                    # Particle classes of the particles in contact
                    class1 = int(M_particle_steady[overlaps[0], 7])
                    class2 = int(M_particle_steady[overlaps[1], 7])
                    # The classes of the particles in contact are assigned to the considered particle
                    M_particle_steady[len(M_particle_steady) - 1, 8 + class1] = M_particle_steady[len(M_particle_steady) - 1, 8 + class1] + 1
                    M_particle_steady[len(M_particle_steady) - 1, 8 + class2] = M_particle_steady[len(M_particle_steady) - 1, 8 + class2] + 1
                    # The class of the considered particle is assigned to the particles in contact 
                    M_particle_steady[overlaps[0], 8 + class_i] = M_particle_steady[overlaps[0], 8 + class_i] + 1
                    M_particle_steady[overlaps[1], 8 + class_i] = M_particle_steady[overlaps[1], 8 + class_i] + 1
                elif n_overlaps_final == 3:
                    # Particle classes of the particles in contact
                    class1 = int(M_particle_steady[overlaps[0], 7])
                    class2 = int(M_particle_steady[overlaps[1], 7])
                    class3 = int(M_particle_steady[overlaps[2], 7])
                    # The classes of the particles in contact are assigned to the considered particle
                    M_particle_steady[len(M_particle_steady) - 1, 8 + class1] = M_particle_steady[len(M_particle_steady) - 1, 8 + class1] + 1
                    M_particle_steady[len(M_particle_steady) - 1, 8 + class2] = M_particle_steady[len(M_particle_steady) - 1, 8 + class2] + 1
                    M_particle_steady[len(M_particle_steady) - 1, 8 + class3] = M_particle_steady[len(M_particle_steady) - 1, 8 + class3] + 1
                    # The class of the considered particle is assigned to the particles in contact 
                    M_particle_steady[overlaps[0], 8 + class_i] = M_particle_steady[overlaps[0], 8 + class_i] + 1
                    M_particle_steady[overlaps[1], 8 + class_i] = M_particle_steady[overlaps[1], 8 + class_i] + 1
                    M_particle_steady[overlaps[2], 8 + class_i] = M_particle_steady[overlaps[2], 8 + class_i] + 1
                elif n_overlaps_final == 4:
                    # Particle classes of the particles in contact
                    class1 = int(M_particle_steady[overlaps[0], 7])
                    class2 = int(M_particle_steady[overlaps[1], 7])
                    class3 = int(M_particle_steady[overlaps[2], 7])
                    class4 = int(M_particle_steady[overlaps[3], 7])
                    # The classes of the particles in contact are assigned to the considered particle
                    M_particle_steady[len(M_particle_steady) - 1, 8 + class1] = M_particle_steady[len(M_particle_steady) - 1, 8 + class1] + 1
                    M_particle_steady[len(M_particle_steady) - 1, 8 + class2] = M_particle_steady[len(M_particle_steady) - 1, 8 + class2] + 1
                    M_particle_steady[len(M_particle_steady) - 1, 8 + class3] = M_particle_steady[len(M_particle_steady) - 1, 8 + class3] + 1
                    M_particle_steady[len(M_particle_steady) - 1, 8 + class4] = M_particle_steady[len(M_particle_steady) - 1, 8 + class4] + 1
                    # The class of the considered particle is assigned to the particles in contact 
                    M_particle_steady[overlaps[0], 8 + class_i] = M_particle_steady[overlaps[0], 8 + class_i] + 1
                    M_particle_steady[overlaps[1], 8 + class_i] = M_particle_steady[overlaps[1], 8 + class_i] + 1
                    M_particle_steady[overlaps[2], 8 + class_i] = M_particle_steady[overlaps[2], 8 + class_i] + 1
                    M_particle_steady[overlaps[3], 8 + class_i] = M_particle_steady[overlaps[3], 8 + class_i] + 1   
                # Update the progress bar
                pbar.update(1)
    # Find NaN values in the steady state array 
    del_list = np.argwhere(np.isnan(M_particle_steady[:, 1]))
    print('Number of particles that could not be placed:', len(del_list))
    print('Diameter of particles that could not be placed:', M_particle_steady[del_list, 5])
    # Delete NaN values from the steady state array 
    M_particle_steady = np.delete(M_particle_steady, del_list, 0)
    # Number of set particles
    n_particles_placed = len(M_particle_steady)
    """ Check the whole particle packing for particles outside the simulation room boarders. """
    # Mask the values between the bounds in x direction
    mask_x = (M_particle_steady[:, 3] <= config.A_0) & (M_particle_steady[:, 3] >= -config.A_0)
    # Mask the values between the bounds in y direction
    mask_y = (M_particle_steady[:, 2] <= config.A_0) & (M_particle_steady[:, 2] >= -config.A_0)
    # Combine the results of both tests 
    mask_xy = [False if not mask_x[i] else mask_y[i] for i in range(len(mask_x))]
    # Extract particles within the compartiment
    M_particle_filtered = M_particle_steady[mask_xy]
    # Number of particles after test
    n_particles_filtered = len(M_particle_filtered)
    # Difference between before and after test
    n_particles_delta = n_particles_placed - n_particles_filtered
    print('Number of particles that were placed outside simulation boarders (deleted):', n_particles_delta)
    """ Check the whole particle packing for overlappings. """
    if config.final_test == True:
        # Counting variable for overlaps
        n_overlaps_total = 0
        # Consider each particles in the packing
        for i in range(len(M_particle_filtered)):
            # Skip NaN values
            if np.isnan(M_particle_filtered[i, 1]):
                continue
            # Retrieve information from the steady state array
            # Coordinates
            z1 = float(M_particle_filtered[i, 1])
            y1 = float(M_particle_filtered[i, 2])
            x1 = float(M_particle_filtered[i, 3])
            # Diameter
            d1 = float(M_particle_filtered[i, 5])
            # Define particle under consideration
            particle1 = z1, y1, x1, d1
            """ Define region of interrest """
            z_min = z1 - test_range
            if z_min < 0:
                z_min = 0
            z_max = z1 + test_range
            y_min = y1 - test_range
            y_max = y1 + test_range
            x_min = x1 - test_range
            x_max = x1 + test_range
            """ Checking particles within the testing range for possible overlappings """
            # Arrays in x, y and z direction with indices of particles, which could overlap with the considered particle
            x_array = np.argwhere(np.logical_and(M_particle_filtered[:, 3] > x_min, M_particle_filtered[:, 3] < x_max))
            y_array = np.argwhere(np.logical_and(M_particle_filtered[:, 2] > y_min, M_particle_filtered[:, 2] < y_max))
            z_array = np.argwhere(np.logical_and(M_particle_filtered[:, 1] > z_min, M_particle_filtered[:, 1] < z_max))
            # Combine the arrays --> Particles to check
            particle_to_check = reduce(np.intersect1d, (x_array, y_array, z_array))
            # Convert array to list
            particle_to_check = np.array(particle_to_check).tolist()
            """ Check each particle that may overlap for an actual overlapping """
            for j in range(len(particle_to_check)):
                if i != j:
                    # Retrieve information from the steady state array
                    # Coordinates
                    z2 = float(M_particle_filtered[j, 1])
                    y2 = float(M_particle_filtered[j, 2])
                    x2 = float(M_particle_filtered[j, 3])
                    # Diameter
                    d2 = float(M_particle_filtered[j, 5])
                    # Define particle
                    particle2 = z2, y2, x2, d2
                    # Required min distance between particles
                    dist_required = (d1 + d2) / 2
                    # Actual distance between particles
                    dist_actual = ((x2 - x1) ** 2 + (y2 - y1) ** 2 + (z2 - z1) ** 2) ** 0.5
                    # If actual distanze is below the min distance, they overlap
                    if (dist_required - dist_actual) > 1e-9:
                        # Update counting variable
                        n_overlaps_total = n_overlaps_total + 1
        # Since every particle is checked with every particle, the overlaps are counted double 
        n_overlaps_total = int(n_overlaps_total / 2)
        print('Total overlaps after final check:', n_overlaps_total)
    print('Random loose packing simulation time:')
    print("--- %s seconds ---" % (time.time() - start_time_rlp))
    # Time stamp
    start_time_growing = time.time()
    """ Delete upper part of particle packing with lower hold-up than average """
    # z-coordinate of the top particle
    z_particle_max = np.max(M_particle_filtered[:, 1])
    # Number of testing compartiments
    n_compartiments = int(round(z_particle_max / (D_32/2), 0))
    # Testing compartment height
    delta_z = z_particle_max / n_compartiments
    # Volume of testing compartiment
    volume_compartiment = delta_z * L_0**2
    # Predefine a list to calculate compartiment holdups
    holdup_compartiments = np.zeros(n_compartiments)
    # Iterate through each comartiment and calculate its holdup
    for i in range(n_compartiments):
        # Mask the values between the defined compartiment bounds
        mask_holdup = (M_particle_filtered[:, 1] <= ((n_compartiments-i) * delta_z)) & (M_particle_filtered[:, 1] >= ((n_compartiments-i-1) * delta_z))
        # Extract particles within the compartiment
        M_particle_compartiment = M_particle_filtered[mask_holdup]
        # Volume of particles within the compartiment
        volume_particles_holdup = np.sum((np.pi / 6) * M_particle_compartiment[:, 5]**3)
        # Compartiment holdup 
        holdup_compartiment_i = volume_particles_holdup / volume_compartiment
        # Update list of compartiment holdups
        holdup_compartiments[i] = holdup_compartiment_i
    # Upper boandary of average holdup calculation
    bound_up_holdup = int(round(n_compartiments * 0.3, 0))
    # Average holdup in packing
    holdup_average = np.mean(holdup_compartiments[bound_up_holdup:n_compartiments-1])
    for holdup_boandary in range(n_compartiments):
        if abs(holdup_average - holdup_compartiments[holdup_boandary]) < 0.05:
            break
    # Max. compartiment height
    z_compartiment_max = (n_compartiments - holdup_boandary) * delta_z
    # Extract final particles 
    M_particle_final = M_particle_filtered[M_particle_filtered[:, 1] <= z_compartiment_max] 
    # Final height of packing
    z_packing = max(M_particle_final[:,1] + M_particle_final[:,5] / 2)
    # Final number of particles in packing
    n_particle_final_rlp = len(M_particle_final)
    print('Final number of particles after deleting the top part of the packing with low holdup:', n_particle_final_rlp)
    """ Graphically display particles in 3D """
    if config.display_particles == True:
        plot_particles(M_particle_final)
    """ 
    Tesselation.
    From here the coordinates order is x,y,z since the tesselation package uses this order.
    Perform Tesselation and get information from Tesselation.
    The cells are polyhedrons.
    The lists are defined as follows:
    volumes_cells: cell volumes
    cell_neighbors: cell neighbors 
    cell_face_normals: normal vectors of cell faces
    cell_vertices: cell verticies
    vertices_indices: indices of the vertices of each cell face
    """
    sys.stdout.write("Tesselation...")
    # Starting coordinates for Tesselation, x, y, z
    tess_start = (-config.A_0, -config.A_0, 0)
    # Final coordinates for Tesselation, x, y, z
    tess_end = (config.A_0, config.A_0, z_packing)
    # Create an array with coordinates for tesselation
    M_particle_tess = M_particle_final[:, [3, 2, 1]]
    # Initial particle diameters
    diameters_0 = M_particle_final[:,5]
    # Calculate tesselation
    volumes_cells, cell_neighbors, cell_face_normals, cell_vertices, vertices_indices = voro_tess(M_particle_tess, tess_start, tess_end, diameters_0)
    sys.stdout.write("completed\n")
    # Convert list to array
    volumes_cells = np.array(volumes_cells)
    """ Calculate initial/max distances between particle center and polyhedron faces
    and determine cell faces which are from beginning in contact with other cells due to initial particle contact """
    sys.stdout.write("Determining initial particle conditions...")
    # Predefine list for initial/max distances between particle center and polyhedron faces
    distances_faces_max = [
        [
            [None for _ in sub_inner_tuple]
            for sub_inner_tuple in inner_tuple
        ]
        for inner_tuple in cell_face_normals
    ]
    # Predefine list for cell faces which are from beginning in contact with other cells due to initial particle contact
    cell_face_fixed = [
        [
            [None for _ in sub_inner_tuple]
            for sub_inner_tuple in inner_tuple
        ]
        for inner_tuple in cell_face_normals
    ]
    # Check each particle from particles array
    for particle_i in range(len(M_particle_tess)):
        # List of fixed cell faces 
        cell_face_fixed_i = []
        # List of distances between particle center and faces at each iteration step for volume grow
        distances_faces_max_i = []
        # Particle center coordinates
        center_particle_i = M_particle_tess[particle_i]
        # Initial particle radius
        radius_particle_i_0 = diameters_0[particle_i] / 2
        # Verticies of the cell corresponding to the particle
        cell_vertices_i = cell_vertices[particle_i]
        # Indices of the vertices of each cell face corresponding to the particle.
        vertices_indices_i = vertices_indices[particle_i]
        # Normal vectors of cell faces corresponding to the particle
        cell_face_normals_i = cell_face_normals[particle_i]
        # Iterate through each face of the cell
        for face in range(len(cell_face_normals_i)):
            # Indices of the vertices of the considered cell face 
            indices_cell_face = vertices_indices_i[face]
            # Vector between the particle center and one point on the face area (vertex)
            vector_center_vertex = center_particle_i - cell_vertices_i[indices_cell_face[0]]
            # Distance between particle/cell center and the cell face
            distance_face = np.abs(np.dot(vector_center_vertex, cell_face_normals_i[face]) / np.linalg.norm(cell_face_normals_i[face]))
            # If the distance is less than or equal to the radius of the sphere, the face intersects with the sphere. 
            # Thus it is set to fixed
            if np.abs(np.abs(distance_face) - radius_particle_i_0) < 1e-9:
                face_fixed = True
            else: 
                face_fixed = False
            # Update lists
            distances_faces_max_i.append(distance_face)
            cell_face_fixed_i.append(face_fixed)
        # Update lists
        distances_faces_max[particle_i] = distances_faces_max_i
        cell_face_fixed[particle_i] = cell_face_fixed_i
    """ Calculate the contact probability of the initial state """
    # Create a new list with cell neighbor diameters of each particle
    # None values are walls
    diameters_neighbors = [
        [diameters_0[i] if i >= 0 else None for i in sublist]
        for sublist in cell_neighbors
    ]    
    # List to label particles. Label is False if particle is in contact with walls at its initial state 
    # or if the solver found no solution during the growing simulation
    particle_status_0 = list([True] * len(diameters_0))
    # Check conditions and update particle_status list accordingly
    # If None values (walls) of the diameters_neighbors list are at the same position as True values (fixed faces) of the
    # cell_face_fixed list, the particle are in contact with walls at their initial state. Thus their label is set to False
    for i in range(len(particle_status_0)):
        for j in range(len(diameters_neighbors[i])):
            if diameters_neighbors[i][j] is None and cell_face_fixed[i][j]:
                particle_status_0[i] = False
    # Number of particles not in contact with walls 
    n_particle_nowall = int(np.sum(particle_status_0))
    # Volume of cells not in contact with walls
    volumes_cells_nowall = volumes_cells[particle_status_0]
    # Initial diameters of particles not in contact with walls
    diameters_nowall_0 = diameters_0[particle_status_0]
    # Initial volumes of particles not in contact with walls
    volumes_particles_nowall_0 = (np.pi / 6) * diameters_nowall_0 **(3)
    # Grow factor. This factor is defined by the lowest cell volume to particle volume ratio 
    # since the volume of each particle can be increased max. to this value due to the required constant size distribution
    # Only particles/cells not in contact with walls are considered
    # Calculate the grow factor only where the particle_status_0 is True
    grow_factor_limit = round(np.min((volumes_cells_nowall / volumes_particles_nowall_0) ** (1/3)), 2)
    grow_factor_max = round(np.max((volumes_cells_nowall / volumes_particles_nowall_0) ** (1/3)), 2)
    if grow_factor_limit <= 1.0:
        print('Fatal error: Tesselation failed. Try another seed')
        sys.exit()
    # Volume of cube surrounding the packing
    volume_cube = z_packing * L_0**2
    # Final particle diameters at constant growing
    diameters_final = diameters_0 * grow_factor_limit
    # Final particle volumes at constant growing
    volumes_particles_final = (np.pi / 6) * diameters_final **(3)
    # Correct volumes
    for i in range(len(volumes_particles_final)):
        if volumes_particles_final[i] - volumes_cells[i] > 1e-9:
            volumes_particles_final[i] = volumes_cells[i]
    # Packing final holdup at constant growing
    holdup_final = np.sum(volumes_particles_final) / volume_cube
    # Starting holdup for grow simulation
    holdup_start = 0.525
    # Grow steps at constant dsd size increase
    grow_steps_const = int(math.floor((holdup_final - holdup_start) / config.holdup_delta_const))
    # Grow steps at non-constant dsd size increase
    grow_steps_nonconst = int(math.floor((1.0 - holdup_final) / config.holdup_delta_nonconst))
    # Total grow steps
    grow_steps = grow_steps_const + grow_steps_nonconst
    # Predefine list with grow factors
    grow_factor_list=[]
    # Grow factor the simulations starts from
    for i in range(grow_steps+1):
        if i <= grow_steps_const:
            holdup_end = holdup_start + config.holdup_delta_const * i
        else:
            holdup_end = 1.0 - (grow_steps + 1 - i) * config.holdup_delta_nonconst
        # Initial guess for solver
        initial_guess = grow_factor_limit
        # Bounds
        bnds = [(1, grow_factor_max)]
        # Solver method
        solver_method = 'SLSQP'
        # Perform the optimization
        result = minimize(solver_grow_factor, initial_guess, args=(holdup_end, volume_cube, diameters_0, volumes_cells), method=solver_method, bounds=bnds, tol=1e-5)
        # Obtain the value of the minimized function
        minimized_value = result.fun
        # Obtain grow factor value
        grow_factor = result.x[0]
        grow_factor_list.append(grow_factor)
    # Predefine list of contact probabilities
    contact_probabilities = [[[[] for _ in range(config.n_classes)] for _ in range(config.n_classes)] for _ in range(grow_steps+4)]
    # Predefine list of average contacts
    average_contacts = [[None] * config.n_classes for _ in range(grow_steps+4)]
    # Determine average contacts and contact probabilities of initial state and update the list
    average_contacts, contact_probabilities = calc_contact_probabilities(average_contacts, contact_probabilities, cell_face_fixed, diameters_neighbors, diameters_0, 0, particle_status_0)
    # Set all faces to fixed which is the case in final state
    cell_face_fixed_final = [[True for _ in row] for row in cell_face_fixed]
    # Determine average contacts and contact probabilities of final state and update the list
    average_contacts, contact_probabilities = calc_contact_probabilities(average_contacts, contact_probabilities, cell_face_fixed_final, diameters_neighbors, diameters_0, grow_steps+3, particle_status_0)
    # Final distribution after random loose packing simulation
    q0_rlp, q3_rlp = calc_q(diameters_nowall_0, n_particle_nowall, d_classes, delta_class)
    # Equivalent particle diameters from cell volumes
    diameters_cells = ((6/np.pi) * volumes_cells)**(1/3)
    # Diameters of droplets not in contact with walls
    diameters_cells_nowall = diameters_cells[particle_status_0]
    # Max. diameter
    d_cells_max = np.max(diameters_cells_nowall)
    # Update distribution classes
    if d_cells_max > d_95:
        # Final number of classes 
        n_classes_final = math.ceil((d_cells_max - d_05) / delta_class)
        # Final class diameters
        d_classes_final = np.array([d_05+delta_class/2 + i * delta_class for i in range(n_classes_final)])
    else:
        d_classes_final = d_classes
    # Final distribution after growing simulation
    q0_final, q3_final = calc_q(diameters_cells_nowall, n_particle_nowall, d_classes_final, delta_class)
    # Particle initial volumes
    volumes_particles = (np.pi / 6) * diameters_0 **(3)
    # Total particle volume
    volume_particle_total = np.sum(volumes_particles)
    # Predefine list of packing holdups
    holdup_packing = np.zeros(grow_steps+4)
    # Packing initial holdup
    holdup_packing[0] = volume_particle_total / volume_cube
    # Packing final holdup
    holdup_packing[grow_steps+3] = 1.0
    #
    sys.stdout.write("completed\n")
    # List to label particles. Label is False if particle's volume is smaller than its cell volume.
    cell_status = list([False] * len(diameters_0))
    # For loop to simulate growing
    for i_grow in range(grow_steps + 2):
        print('Particle grow simulation. Step:', i_grow+1, 'of', (grow_steps + 2))
        # Increase diameters
        if i_grow <= grow_steps_const:
            diameters = diameters_0 * grow_factor_list[i_grow]
        elif i_grow == (grow_steps_const + 1):
            diameters = diameters_0 * grow_factor_limit
        else:
            diameters = diameters_0 * grow_factor_list[i_grow-1]
        # Update volumes
        volumes_particles = (np.pi / 6) * diameters **(3)
        # Check and update cell status 
        for i in range(len(cell_status)):
            if volumes_particles[i] - volumes_cells[i] > 1e-9:
                volumes_particles[i] = volumes_cells[i]
                cell_status[i] = True
        # Final total particle volume
        volume_particle_total = np.sum(volumes_particles)
        # Update packing holdup
        holdup_packing[i_grow+1] = volume_particle_total / volume_cube
        # List of numbers to be squared
        particle_indices = list(range(0, len(M_particle_tess)))
        sys.stdout.write("Initializing multiprocessing...")
        # Initialize manager for shared data
        with multiprocessing.Manager() as manager:
            # Lists shared with manager
            cell_face_fixed = manager.list([manager.list(sublist) for sublist in cell_face_fixed])
            particle_status = manager.list(particle_status_0)
            # Number of processes to use
            if config.n_cores == 0:
                num_processes = multiprocessing.cpu_count()
            else:
                num_processes = config.n_cores
            # Create a pool of worker processes
            with multiprocessing.Pool(processes=num_processes) as pool:
                # Create partial function for parallel processing
                calc_partial = partial(calc_grow,
                                       config=config,
                                       volumes_particles=volumes_particles,
                                       diameters=diameters,
                                       cell_face_fixed=cell_face_fixed,
                                       distances_faces_max=distances_faces_max,
                                       cell_face_normals=cell_face_normals,
                                       vertices_indices=vertices_indices,
                                       cell_vertices=cell_vertices,
                                       particle_status=particle_status,
                                       cell_status=cell_status)
                # List to keep track of AsyncResult objects
                async_results = []
                # Distribute the tasks among the worker processes using apply_async
                for index in particle_indices:
                    result = pool.apply_async(calc_partial, (index,))
                    async_results.append(result)
                sys.stdout.write("completed\n")
                sys.stdout.write("Particle grow simulation process...")
                """# Tracker does not work since the script has been modulized:
                # Use tqdm to track progress
                with tqdm(total=len(async_results), desc='Particle grow simulation progress', unit='particle') as pbar:
                    results = []
                    for result in async_results:
                        results.append(result.get())
                        # Update progress bar after each task completes
                        pbar.update(1)"""
            # No need to close the pool and wait, 'with' statement handles it
            # Convert shared lists to regular lists for further processing 
            cell_face_fixed = [list(sublist) for sublist in cell_face_fixed]
            particle_status = list(particle_status)
            sys.stdout.write("completed\n")
        # Determine contact probabilities and update the list
        average_contacts, contact_probabilities = calc_contact_probabilities(average_contacts, contact_probabilities, cell_face_fixed, diameters_neighbors, diameters_0, i_grow+1, particle_status)
    # Final number of particles considered for the growing simulation
    n_particle_final_gs = int(np.sum(particle_status))
    """ Save results as Excel sheet. """
    save_excel(config, ROOT_DIR, NAME_EXCEL_FILE, holdup_packing, average_contacts, contact_probabilities, q0_initial, q0_rlp, q0_final, q3_initial, q3_rlp, q3_final, d_classes_final, n_particle_final_rlp, n_particle_nowall, n_particle_final_gs)
    """ Graphically display particles in 3D """
    if config.display_particles == True:
        plot_particles(M_particle_final)
    print('Particle growing simulation time:')
    print("--- %s seconds ---" % (time.time() - start_time_growing))
    print('Total simulation time:')
    print("--- %s seconds ---" % (time.time() - start_time_rlp))
